from rdflib import Graph, Namespace, RDF, RDFS, OWL, term
from collections import defaultdict, deque

import networkx as nx
import os
from tqdm import tqdm
import pandas as pd
from pyjedai.datamodel import Data
from pyjedai.block_building import StandardBlocking
from pyjedai.block_cleaning import BlockPurging
from pyjedai.block_cleaning import BlockFiltering
from pyjedai.matching import EntityMatching
from pyjedai.comparison_cleaning import WeightedEdgePruning
from pyjedai.clustering import ConnectedComponentsClustering


class KnowledgeGraph:
    def __init__(self, graph_file, ontology_file):
        self.graph = Graph()
        self.ontology = Graph()
        self.triple_dict = {}

        # Load N-Triples formatted files
        self.graph.parse(graph_file)
        self.ontology.parse(ontology_file)

    def create_clusters(self,prune_top_nodes=6,max_cluster_size=500):
        # convert RDFLib graph to Networkx MultiDiGraph
        nx_graph = nx.MultiDiGraph()

        for s, p, o in tqdm(self.graph):
            self.__add_triple(s,p,o)
            if p != RDF.type:
                if o.__class__ == term.Literal:
                    continue  
                nx_graph.add_edge(s, o, **{'key': p})

        # get the number of nodes, edges, and self-loops
        nodes = nx.number_of_nodes(nx_graph)
        edges = nx.number_of_edges(nx_graph)
        self_loops = nx.number_of_selfloops(nx_graph)

        print('There are {} nodes, {} edges, and {} self-loop(s)'.format(nodes, edges, self_loops))

        # get connected components -- have to convert MultiDiGraph to undirected graph
        nx_graph_und = nx_graph.to_undirected()

        # get connected components
        top_k_nodes = self.get_top_k_by_degree(nx_graph_und,prune_top_nodes)
        for node in top_k_nodes:
            nx_graph_und.remove_node(node[0])


        components = sorted(list(nx.connected_components(nx_graph_und)), key=len, reverse=True)

        new_components=[]
        for component in components:
            if len(component) > max_cluster_size:
                dataList = []
                for node in component:
                    edges = self.triple_dict.get(node,-1)
                    if edges != -1:
                        for (p,o) in edges:
                            if o.__class__ != term.URIRef:
                                dataList.append((node,p,o))
                df = pd.DataFrame(dataList, columns =['s', 'p', 'o'])

                # Create pyjedai data format
                attr=['s','p','o']
                data = Data(
                    dataset_1=df,
                    id_column_name_1='s',
                    attributes_1=attr,
                )

                # Build blocks from connected component
                bb = StandardBlocking()
                blocks = bb.build_blocks(data)

                # Block cleaning
                bp = BlockPurging()
                cleaned_blocks = bp.process(blocks, data, tqdm_disable=False)
                bc = BlockFiltering(ratio=0.75)
                cleaned_blocks = bc.process(cleaned_blocks, data)

                mb = WeightedEdgePruning(weighting_scheme='CBS')
                blocks = mb.process(cleaned_blocks, data)

                em = EntityMatching(
                    metric='jaccard',
                    similarity_threshold=0.4
                )

                pairs_graph = em.predict(blocks, data)

                ec = ConnectedComponentsClustering()
                clusters = ec.process(pairs_graph, data, similarity_threshold=0.3)

                
                for idx, cluster in enumerate(clusters):
                    c = set()
                    for item in cluster:
                        c.add(df.iloc[item]["s"])
                    new_components.append(list(c))

        components.extend(new_components)

        #Final cleaning steps for components - remove componenets with less than two literal properties
        filtered_components = []

        for component in components:
            new_component = []
            for node in component:
                edges = self.triple_dict.get(node, [])
                literal_count = sum(1 for _, o in edges if isinstance(o, term.Literal))
                if literal_count >= 2:
                    new_component.append(node)
            if new_component:
                filtered_components.append(new_component)

        # Optionally replace original components list
        components = filtered_components

        return components, self.triple_dict

    def get_top_k_by_degree(self, graph, k=10):
        if k==0:
            return []
        # Compute degree for all nodes
        degree_dict = dict(graph.degree())
        
        # Sort nodes by degree in descending order
        top_k = sorted(degree_dict.items(), key=lambda item: item[1], reverse=True)[:k]
        
        return top_k
    
    def __add_triple(self,s,p,o):
        l = self.triple_dict.get(s,-1)
        if l == -1:
            l = []
        l.append((p,o))
        self.triple_dict[s] = l





from rdflib import RDF, term
import csv

def export_data(clusters, triples_dict):
    with open("test_training.csv", "w", newline='', encoding="utf-8") as f:
        writer = csv.writer(f)
        writer.writerow(["triples", "label"])  # CSV header

        for cluster in clusters:
            result = ""
            label = "No"

            for node in cluster:
                edges = triples_dict.get(node, [])
                for p, o in edges:
                    # Build triple string
                    subj = f"<{node}>"
                    pred = f"<{p}>"
                    if isinstance(o, term.Literal):
                        obj = f'"{str(o)}"'
                    else:
                        obj = f"<{o}>"
                    result += f"{subj} {pred} {obj} .\n"

                    # Check if it's a KeyValuePair
                    if obj.find("KeyValue")!=-1:
                        label = "Yes"

            writer.writerow([result.strip(), label])



import os
import rdflib
from rdflib.term import URIRef, Literal
from gilgamesh_summarizer.kvpair import KvPair

class Rewriter:

    def __init__(self, input_file, output_file, output_csv, uri, resource, summarized_triples, removedLinks):
        self.input_file = input_file
        self.output_file = output_file
        self.output_csv = output_csv
        self.key_value_map = {}
        self.metrics_map = {}
        self.temporal_attributes = {}
        self.URI = uri
        self.RESOURCE = resource
        self.summarized_triples = summarized_triples
        self.removedLinks = []
        for link in removedLinks:
            self.removedLinks.append(str(link))

    def parse(self, export_to_csv=False):
        self.collect_entities()

        graph = rdflib.Graph()
        graph.parse(self.input_file, format='nt')

        with open(self.output_file, 'w') as writer:
            csv_writer = open(self.output_csv, 'w') if export_to_csv else None

            for type, keyString, valueString in self.summarized_triples:
                for s, p, o in graph:
                    s_str, p_str, o_str = str(s), str(p), str(o)

                    if str(type) in o_str:
                        continue

                    if p_str == str(keyString):
                        self.set_kv_key(s_str, o_str.strip('"'))
                        continue
                    if p_str == str(valueString):
                        self.set_kv_value(s_str, o_str)
                        continue
                    if any(kw in p_str for kw in self.removedLinks):
                        self.set_kv_subject(o_str, s_str)
                        self.set_kv_old_predicate(o_str, p_str)
                        continue
                    if self.key_value_map.get(s_str) and p_str == "http://purl.org/dc/terms/issued":
                        self.metrics_map[s_str] = o_str
                        continue
                    if p_str == "http://purl.org/dc/terms/identifier" and o_str.strip('"') in s_str:
                        ostr = o_str.strip('\"')
                        writer.write(f"{s_str} {p_str} <{self.RESOURCE}{ostr}> ." "\n")
                        continue
                    if p_str == "http://xmlns.com/foaf/0.1/accountName":
                        identifier = o_str.strip('"')
                        writer.write(f"{s_str} {p_str} <{self.RESOURCE}{identifier}> ." "\n")
                        writer.write(f"{s_str} <http://purl.org/dc/terms/identifier> <{self.RESOURCE}{identifier}> ." "\n")
                        continue
                    if p_str == "http://purl.org/dc/terms/spatial":
                        if 'POLYGON((' in o_str:
                            geom_uri = f"<{s_str.strip('<>')}/geometry>"
                            writer.write(f"{s_str} <http://www.opengis.net/ont/geosparql#hasGeometry> {geom_uri} ." "\n")
                            writer.write(f"{geom_uri} <http://www.opengis.net/ont/geosparql#asWKT> {o_str} ." "\n")
                        else:
                            p_str = "http://www.opengis.net/ont/geosparql#hasGeometry"
                    elif p_str == "http://www.w3.org/ns/locn#geometry":
                        p_str = "http://www.opengis.net/ont/geosparql#asWKT"

                    writer.write(f"{s_str} {p_str} {o_str} ." "\n")

                props = set()

            for key, kv in self.key_value_map.items():
                if export_to_csv and key not in self.metrics_map and kv.getSubject():
                    csv_writer.write(kv.toCSVEntry(key) + "\n")
                if kv.getSubject():
                    writer.write(kv.getTriple())
                    props.add(kv.getDataProperty())

            for key, issued_val in self.metrics_map.items():
                kv = self.key_value_map.get(key)
                if kv and kv.getSubject():
                    if export_to_csv:
                        csv_writer.write(kv.toCsvEntryExtraProperties(key, [("http://purl.org/dc/terms/issued", issued_val)]) + "\n")
                    writer.write(f"{kv.getSubject()} <{self.URI}metricIssued> {issued_val} ." "\n")

            for prop in props:
                print(prop)

            if csv_writer:
                csv_writer.close()

    def collect_entities(self):
        graph = rdflib.Graph()
        graph.parse(self.input_file, format='nt')

        for s, _, o in graph:
            s_str, o_str = str(s), str(o)
            if 'KeyValuePair' in o_str:
                self.key_value_map[s_str] = KvPair()

    def set_kv_key(self, key, kv_key):
        if key in self.key_value_map:
            self.key_value_map[key].setKey(kv_key)

    def set_kv_value(self, key, kv_value):
        if key in self.key_value_map:
            self.key_value_map[key].setValue(kv_value)

    def set_kv_subject(self, key, subject):
        if key not in self.key_value_map:
            self.key_value_map[key] = KvPair()
        self.key_value_map[key].setSubject(subject)

    def set_kv_old_predicate(self, key, old_predicate):
        if key in self.key_value_map:
            self.key_value_map[key].setOldPredicate(old_predicate)

    def get_kv_subject(self, key):
        return self.key_value_map.get(key).getSubject()
